@@ -25,6 +25,6 @@ For further information on X2Go, please visit the X2Go wiki:
 http://wiki.x2go.org
 
 """
-__VERSION__ = "0.5.0.4"
+__VERSION__ = "0.6.0.1"
 
-from frontend import *
+from .frontend import *
