@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright (C) 2010-2016 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
+# Copyright (C) 2010-2018 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
 #
 # PyHoca CLI is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License as published by
@@ -25,9 +25,8 @@ import x2go
 import sys
 import os
 import copy
-import time
 import getpass
-from gevent import socket
+from gevent import socket, sleep
 
 # for debugging
 import pprint
@@ -70,6 +69,14 @@ class PyHocaCLI(x2go.X2GoClient):
     def _runtime_error(self, m, exitcode=-1):
         runtime_error(m, exitcode=exitcode)
 
+    def HOOK_no_such_command(self, *args, **kwargs):
+        x2go.X2GoClient.HOOK_no_such_command(self, *args, **kwargs)
+        if len(args) == 0:
+            cmd=kwargs['cmd']
+        else:
+            cmd=args[0]
+        self._runtime_error("The remote X2Go Server does not know how to execute the command '{cmd}'. Aborting...".format(cmd=cmd))
+
     def list_sessions(self, s_hash):
         """\
         List up server-side available sessions for the logged in user.
@@ -81,35 +88,35 @@ class PyHocaCLI(x2go.X2GoClient):
 
         """
         # retrieve a session list
-        print
-        print "Available runing/suspended X2Go sessions"
-        print "========================================"
+        print()
+        print("Available runing/suspended X2Go sessions")
+        print("========================================")
         _peername = self._X2GoClient__get_session_server_peername(s_hash)
-        print "Host: %s - [%s]:%s" % (self._X2GoClient__get_session_server_hostname(s_hash), _peername[0], _peername[1])
-        print "Username: %s" % self._X2GoClient__get_session_username(s_hash)
-        print
+        print("Host: %s - [%s]:%s" % (self._X2GoClient__get_session_server_hostname(s_hash), _peername[0], _peername[1]))
+        print("Username: %s" % self._X2GoClient__get_session_username(s_hash))
+        print()
         session_infos = self._X2GoClient__list_sessions(s_hash)
         if session_infos:
-            for session_info in session_infos.values():
-                print "Session Name: %s" % session_info
-                print "-------------"
-                print "cookie: %s" % session_info.cookie
-                print "agent PID: %s" % session_info.agent_pid
-                print "display: %s" % session_info.display
-                print "status: %s" % session_info.status
-                print "graphic port: %s" % session_info.graphics_port
-                print "snd port: %s" % session_info.snd_port
-                print "sshfs port: %s" % session_info.sshfs_port
-                print "username: %s" % session_info.username
-                print "hostname: %s" % session_info.hostname
+            for session_info in list(session_infos.values()):
+                print("Session Name: %s" % session_info)
+                print("-------------")
+                print("cookie: %s" % session_info.cookie)
+                print("agent PID: %s" % session_info.agent_pid)
+                print("display: %s" % session_info.display)
+                print("status: %s" % session_info.status)
+                print("graphic port: %s" % session_info.graphics_port)
+                print("snd port: %s" % session_info.snd_port)
+                print("sshfs port: %s" % session_info.sshfs_port)
+                print("username: %s" % session_info.username)
+                print("hostname: %s" % session_info.hostname)
                 # TODO: turn into datetime object
-                print "create date: %s" % session_info.date_created
+                print("create date: %s" % session_info.date_created)
                 # TODO: turn into datetime object
-                print "suspended since: %s" % session_info.date_suspended
-                print
+                print("suspended since: %s" % session_info.date_suspended)
+                print()
         else:
-            print "No running/suspended sessions found on X2Go server."
-            print
+            print("No running/suspended sessions found on X2Go server.")
+            print()
 
     def list_desktops(self, s_hash):
         """\
@@ -123,20 +130,20 @@ class PyHocaCLI(x2go.X2GoClient):
 
         """
         # retrieve a desktop list
-        print
-        print "X2Go desktops available for sharing"
-        print "==================================="
+        print()
+        print("X2Go desktops available for sharing")
+        print("===================================")
         _peername = self._X2GoClient__get_session_server_peername(s_hash)
-        print "Host: %s - [%s]:%s" % (self._X2GoClient__get_session_server_hostname(s_hash), _peername[0], _peername[1])
-        print "Username: %s" % self._X2GoClient__get_session_username(s_hash)
-        print
+        print("Host: %s - [%s]:%s" % (self._X2GoClient__get_session_server_hostname(s_hash), _peername[0], _peername[1]))
+        print("Username: %s" % self._X2GoClient__get_session_username(s_hash))
+        print()
         desktop_list = self._X2GoClient__list_desktops(s_hash)
         if desktop_list:
             for desktop_name in desktop_list:
-                print desktop_name
+                print(desktop_name)
         else:
-            print "No X2Go desktop sessions found that are available for desktop sharing."
-            print
+            print("No X2Go desktop sessions found that are available for desktop sharing.")
+            print()
 
     def list_profiles(self):
         """\
@@ -146,16 +153,16 @@ class PyHocaCLI(x2go.X2GoClient):
 
         """
         # retrieve a session list
-        print
-        print "Available X2Go session profiles"
-        print "==============================="
+        print()
+        print("Available X2Go session profiles")
+        print("===============================")
         _profiles = self._X2GoClient__get_profiles()
         for _profile_id in _profiles.profile_ids:
             _profile_config = _profiles.get_profile_config(_profile_id)
             _session_params = _profiles.to_session_params(_profile_id)
-            print _profile_config
-            print _session_params
-            print
+            print(_profile_config)
+            print(_session_params)
+            print()
 
     def clean_sessions(self, s_hash):
         """\
@@ -219,16 +226,37 @@ class PyHocaCLI(x2go.X2GoClient):
         """
         # resume a running session
         self.logger('resuming X2Go session: %s' % self.args.resume, loglevel=x2go.loglevel_INFO, )
-        available_sessions = self._X2GoClient__list_sessions(s_hash)
+        if available_sessions is None:
+            available_sessions = self._X2GoClient__list_sessions(s_hash)
         if available_sessions and self.args.resume == 'OLDEST':
             self._auto_resume_oldest(s_hash)
         elif available_sessions and self.args.resume == 'NEWEST':
             self._auto_resume_newest(s_hash)
-        elif self.args.resume in available_sessions.keys():
+        elif self.args.resume in list(available_sessions.keys()):
             self._X2GoClient__resume_session(s_hash, self.args.resume)
         else:
             self._runtime_error('requested session not available on X2Go server [%s]:%s' % (self.args.server, self.args.remote_ssh_port), exitcode=20)
 
+    def try_resume_session(self, s_hash):
+        """\
+        Use C{X2GoSession} object with session UUID C{s_hash} to resume a server-side suspended X2Go session
+        or to take over a server-side running X2Go session.
+
+        @param s_hash: session UUID
+        @type s_hash: C{str}
+
+        """
+        self.logger('trying to resume / take over X2Go session with command: %s' % self.args.command, loglevel=x2go.loglevel_INFO, )
+        available_sessions = self._X2GoClient__list_sessions(s_hash, with_command=self.args.command)
+        if available_sessions:
+            self.logger('matching sessions found: %s' % available_sessions.keys(), loglevel=x2go.loglevel_INFO, )
+            self.logger('picking the first in list: %s' % list(available_sessions.keys())[0], loglevel=x2go.loglevel_INFO, )
+            # FIXME: what session do we want to resume really (oldest, newest, etc.)?
+            self._X2GoClient__resume_session(s_hash, list(available_sessions.keys())[0])
+            return True
+        else:
+            return False
+
     def share_desktop_session(self, s_hash):
         """\
         Use C{X2GoSession} object with session UUID C{s_hash} to share a server-side X2Go desktop session that is available for sharing.
@@ -243,7 +271,7 @@ class PyHocaCLI(x2go.X2GoClient):
         self.logger('sharing X2Go session: %s' % _desktop, loglevel=x2go.loglevel_INFO, )
         try:
             self._X2GoClient__share_desktop_session(s_hash, desktop=_desktop, share_mode=_share_mode)
-        except x2go.X2GoDesktopSharingException, e:
+        except x2go.X2GoDesktopSharingException as e:
             self._runtime_error('%s' % str(e), exitcode=54)
 
 
@@ -265,7 +293,7 @@ class PyHocaCLI(x2go.X2GoClient):
             _session_names = [ self.args.suspend ]
 
         for _session_name in _session_names:
-            if _session_name in available_sessions.keys():
+            if _session_name in list(available_sessions.keys()):
                 self._X2GoClient__suspend_session(s_hash, _session_name)
                 self._pyhoca_logger("X2Go session %s has been suspended" % _session_name, loglevel=x2go.loglevel_NOTICE, )
             else:
@@ -294,7 +322,7 @@ class PyHocaCLI(x2go.X2GoClient):
             _session_names = [ self.args.terminate ]
 
         for _session_name in _session_names:
-            if _session_name in available_sessions.keys():
+            if _session_name in list(available_sessions.keys()):
                 self._X2GoClient__terminate_session(s_hash, _session_name)
                 self._pyhoca_logger("X2Go session %s has been terminated" % _session_name, loglevel=x2go.loglevel_NOTICE, )
             else:
@@ -360,26 +388,26 @@ class PyHocaCLI(x2go.X2GoClient):
 
             _session_profiles = self._X2GoClient__get_profiles()
             # retrieve a session list
-            print
-            print "Available X2Go session profiles"
-            print "==============================="
+            print()
+            print("Available X2Go session profiles")
+            print("===============================")
             if hasattr(_session_profiles, 'config_files') and _session_profiles.config_files is not None:
-                print "configuration files: %s" % _session_profiles.config_files
+                print("configuration files: %s" % _session_profiles.config_files)
             if hasattr(_session_profiles, 'user_config_file') and _session_profiles.user_config_file is not None:
-                print "user configuration file: %s" % _session_profiles.user_config_file
+                print("user configuration file: %s" % _session_profiles.user_config_file)
             if hasattr(_session_profiles, 'broker_url') and _session_profiles.broker_url is not None:
-                print "X2Go Session Broker URL: %s" % _session_profiles.broker_url
-            print
+                print("X2Go Session Broker URL: %s" % _session_profiles.broker_url)
+            print()
             for _profile_id in _session_profiles.profile_ids:
                 _profile_config = _session_profiles.get_profile_config(_profile_id)
                 _session_params = _session_profiles.to_session_params(_profile_id)
-                print 'Profile ID: %s' % _profile_id
-                print 'Profile Name: %s' % _session_params['profile_name']
-                print 'Profile Configuration:'
+                print('Profile ID: %s' % _profile_id)
+                print('Profile Name: %s' % _session_params['profile_name'])
+                print('Profile Configuration:')
                 pprint.pprint(_profile_config)
-                print 'Derived session parameters:'
+                print('Derived session parameters:')
                 pprint.pprint(_session_params)
-                print
+                print()
 
             # done
             sys.exit(0)
@@ -414,6 +442,12 @@ class PyHocaCLI(x2go.X2GoClient):
 
         else:
 
+            xinerama = False
+            if self.args.xinerama:
+                xinerama = self.args.xinerama
+
+            allow_share_local_folders = bool(self.args.share_local_folders)
+
             # setup up the manually configured X2Go session
             self.x2go_session_hash = self._X2GoClient__register_session(args.server, port=int(self.args.remote_ssh_port),
                                                            known_hosts=ssh_known_hosts_filename,
@@ -423,6 +457,7 @@ class PyHocaCLI(x2go.X2GoClient):
                                                            profile_name = 'Pyhoca-Client_Session',
                                                            session_type=self.args.session_type,
                                                            link=self.args.link,
+                                                           xinerama=self.args.xinerama,
                                                            geometry=self.args.geometry,
                                                            pack=self.args.pack,
                                                            cache_type='unix-kde',
@@ -433,6 +468,8 @@ class PyHocaCLI(x2go.X2GoClient):
                                                            print_action=self.args.print_action,
                                                            print_action_args=self.args.print_action_args,
                                                            share_local_folders=self.args.share_local_folders,
+                                                           allow_share_local_folders=allow_share_local_folders,
+                                                           forward_sshagent=self.args.forward_sshagent,
                                                            cmd=self.args.command)
 
     def authenticate(self):
@@ -441,7 +478,23 @@ class PyHocaCLI(x2go.X2GoClient):
 
         """
         connected = False
-        force_password_auth = False
+
+        if self.args.password:
+            password = self.args.password
+            self.args.password = None
+            cmdline_password = True
+            force_password_auth = True
+        else:
+            password = None
+            cmdline_password = False
+            force_password_auth = False
+
+        if self.args.force_password:
+            force_password_auth = True
+
+        passphrase = None
+        passphrase_unlock_counter = 3
+
         _username = self.args.username or self._X2GoClient__get_session_username(self.x2go_session_hash)
 
         # if we still don't have a valid user name, try the broker...
@@ -449,51 +502,171 @@ class PyHocaCLI(x2go.X2GoClient):
             _username = self.session_profiles.get_broker_username()
 
         try:
-            _auth_count = self.auth_attempts +1
+            _auth_count = self.auth_attempts
             while not connected and _auth_count:
+
                 try:
-                    self._X2GoClient__connect_session(self.x2go_session_hash, username=_username, password=self.args.password, force_password_auth=force_password_auth)
+
+                    # decrement authentication counter...
+                    _auth_count -= 1
+
+                    # show interactive password prompt
+                    if force_password_auth and not cmdline_password:
+
+                        password = getpass.getpass()
+
+                        # workaround for Python bug: http://bugs.python.org/issue11236
+                        try:
+                            if password is not None and '\x03' in password:
+                                raise KeyboardInterrupt()
+                        except KeyboardInterrupt:
+                            self._runtime_error('Authentication cancelled by user by hitting Ctrl-C at password prompt', exitcode=-200)
+
+                        if not password:
+                            self._pyhoca_logger('password is empty, please re-try... (hit Ctrl-C plus ENTER to cancel)', loglevel=x2go.loglevel_WARN, )
+                            _auth_count += 1
+                            continue
+
+                    # connection attempt with remote X2Go Server
+                    self._X2GoClient__connect_session(self.x2go_session_hash, username=_username, password=password, passphrase=passphrase, force_password_auth=force_password_auth)
+
+                    # we succeeeded
                     connected = True
-                    force_password_auth = False
-                except x2go.PasswordRequiredException, e:
-                    self._pyhoca_logger('unlock SSH key file (%s)' % self.args.ssh_privkey, loglevel=x2go.loglevel_NOTICE, )
-                    self.args.password = getpass.getpass()
-                except x2go.AuthenticationException, e:
-                    force_password_auth = True
-                    self._pyhoca_logger('passwordless login for ,,%s\'\' failed' % _username, loglevel=x2go.loglevel_WARN, )
-                    self._pyhoca_logger('proceeding to interactive login for user ,,%s\'\'' % _username, loglevel=x2go.loglevel_NOTICE, )
+
+                    # this will end the loop (as connected is set to True)
+                    continue
+
+                except x2go.PasswordRequiredException as e:
+
+                    # x2go.PasswordRequiredException: This exception gets raised if an SSH pubkey is protected by a passphrase
+
+                    if not force_password_auth and passphrase_unlock_counter >= 1:
+                        if passphrase == '':
+                            self._pyhoca_logger('empty SSH key passphrase (%s), try again...' % self.args.ssh_privkey, loglevel=x2go.loglevel_WARN, )
+                        self._pyhoca_logger('unlock SSH key file (%s)' % self.args.ssh_privkey, loglevel=x2go.loglevel_NOTICE, )
+                        passphrase = getpass.getpass('Passphrase: ')
+                        passphrase_unlock_counter -= 1
+                        # undo auth counter decrement
+                        _auth_count += 1
+                        continue
+
+                    if not force_password_auth and _auth_count >= 1:
+                        self._pyhoca_logger('unlocking of SSH key failed, proceeding with interactive login', loglevel=x2go.loglevel_WARN, )
+                        force_password_auth = True
+                        password = None
+                        passphrase = None
+                        _auth_count += 1
+                        continue
+
+                except x2go.AuthenticationException as e:
+
+                    # x2go.AuthenticationException: This exception gets raised if the authentication failed
+
+                    if force_password_auth:
+                        self._pyhoca_logger('password based login for ,,%s\'\' failed [AuthException]' % _username, loglevel=x2go.loglevel_WARN, )
+                    else:
+                        self._pyhoca_logger('passwordless login for ,,%s\'\' failed [AuthException]' % _username, loglevel=x2go.loglevel_WARN, )
+
+                    # if the previous login attempt was pubkey based, enforce interactive login for the next round...
+                    if not password and _auth_count >= 1:
+                        self._pyhoca_logger('proceeding to interactive login for user ,,%s\'\'' % _username, loglevel=x2go.loglevel_NOTICE, )
+                        force_password_auth = True
+                        # undo auth counter decrement
+                        _auth_count += 1
+
+                    # a password was provided via the command line
+                    elif password and cmdline_password and _auth_count >= 1:
+                        self._pyhoca_logger('cmdline provided password failed, proceeding to interactive login for user ,,%s\'\'' % _username, loglevel=x2go.loglevel_WARN, )
+                        force_password_auth = True
+                        cmdline_password = False
+
+                    # else, if the previous attempt was already interactive, offer re-trying
+                    elif force_password_auth and _auth_count >= 1:
+                        self._pyhoca_logger('please re-try login for user ,,%s\'\'' % _username, loglevel=x2go.loglevel_NOTICE, )
+
+                    passphrase = None
+                    password = None
+
                 except x2go.BadHostKeyException:
+
+                    # let's bail out here immediately...
+
                     self._runtime_error('SSH host key verification for remote host [%s]:%s failed' % (self.args.server, self.args.remote_ssh_port), exitcode=-254)
-                except x2go.SSHException, e:
-                    if str(e) not in ('not a valid DSA private key file', 'Incompatible ssh peer (no acceptable kex algorithm)', 'No authentication methods available'):
+
+                except x2go.SSHException as e:
+
+                    # this bit only captures problems with the SSH key file, other
+                    # SSHExceptions are simply ignored (and we proceed to interactive login)
+
+                    if str(e).lower().startswith('could not deserialize key data')                                          \
+                        :
+
+                        # this error gets thrown when the passphrase for unlocking the SSH privkey was wrong
+
+                        if not force_password_auth and passphrase_unlock_counter >= 1:
+
+                            if passphrase is not None and passphrase != '':
+                                self._pyhoca_logger('wrong SSH key passphrase (%s), try again...' % self.args.ssh_privkey, loglevel=x2go.loglevel_WARN, )
+                            self._pyhoca_logger('unlock SSH key file (%s)' % self.args.ssh_privkey, loglevel=x2go.loglevel_NOTICE, )
+                            passphrase = getpass.getpass('Passphrase: ')
+                            passphrase_unlock_counter -= 1
+                            # undo auth counter decrement
+                            _auth_count += 1
+                            continue
+
+                        if not force_password_auth and _auth_count >= 1:
+
+                            self._pyhoca_logger('unlocking of SSH key failed, proceeding with interactive login', loglevel=x2go.loglevel_WARN, )
+                            force_password_auth = True
+                            password = None
+                            passphrase = None
+                            _auth_count += 1
+                            continue
+
+                    elif not str(e).lower().startswith('not a valid dsa private key file') and                        \
+                         not str(e).lower().startswith('not a valid rsa private key file') and                        \
+                         not str(e).lower().startswith('incompatible ssh peer (no acceptable kex algorithm)') and     \
+                         not str(e).lower().startswith('no authentication methods available')                         \
+                        :
+
+                        if force_password_auth:
+
+                            self._pyhoca_logger('password based login for ,,%s\'\' failed [SSHException]' % _username, loglevel=x2go.loglevel_WARN, )
+                        else:
+                            self._pyhoca_logger('passwordless login for ,,%s\'\' failed [SSHException]' % _username, loglevel=x2go.loglevel_WARN, )
+
+                        # let's bail out here...
                         self._runtime_error(str(e), exitcode=253)
-                    force_password_auth = True
-                    self._pyhoca_logger('passwordless login for ,,%s\'\' failed' % _username, loglevel=x2go.loglevel_WARN, )
-                    self._pyhoca_logger('proceeding to interactive login for user ,,%s\'\'' % _username, loglevel=x2go.loglevel_NOTICE, )
 
-                if force_password_auth:
-                    self.args.password = getpass.getpass()
+                    else:
 
-                # workaround for Python bug: http://bugs.python.org/issue11236
-                try:
-                    if self.args.password is not None and '\x03' in self.args.password:
-                        raise KeyboardInterrupt()
-                except KeyboardInterrupt:
-                    self._runtime_error('Authentication cancelled by user by hitting Ctrl-C at password prompt', exitcode=-200)
+                        self._pyhoca_logger('[SSHException] the following error will be ignored: %s' % str(e), loglevel=x2go.loglevel_WARN)
+                        self._pyhoca_logger('proceeding to interactive login for user ,,%s\'\'' % _username, loglevel=x2go.loglevel_NOTICE, )
+                        force_password_auth = True
+                        password = None
+                        passphrase = None
+                        _auth_count += 1
 
-                _auth_count -= 1
-
-                if not connected and not _auth_count:
+                if not connected and _auth_count <= 0:
                     if self.auth_attempts >= 2:
                         self._runtime_error('authentication failed, too many failures during interactive login', exitcode=-201)
                     elif self.auth_attempts == 1:
-                        self._runtime_error('interactive authentication failed', exitcode=-202)
-                    else:
-                        self._runtime_error('non-interactive authentication failed', exitcode=-203)
+                        self._runtime_error('authentication failed', exitcode=-202)
 
-        except socket.error, e:
+
+        except socket.error as e:
             self._runtime_error('a socket error occured while establishing the connection: %s' % str(e), exitcode=-245)
 
+        # drop clear text passwords...
+        password = None
+        passphrase = None
+
+        if not self.is_x2goserver(self.x2go_session_hash):
+            self._runtime_error('the remote server does not have the X2Go Server software installed', exitcode=1337)
+
+        self._pyhoca_logger('authentication has been successful', loglevel=x2go.loglevel_NOTICE, )
+
+
     def MainLoop(self):
         """\
         Start the main loop of this application.
@@ -520,6 +693,9 @@ class PyHocaCLI(x2go.X2GoClient):
                 self.list_profiles()
                 sys.exit(0)
 
+            if self.args.resume or self.args.new:
+                self._pyhoca_logger("give the X2Go session some time to come up...", loglevel=x2go.loglevel_NOTICE, )
+
             if self.args.resume:
                 self.resume_session(self.x2go_session_hash)
 
@@ -532,20 +708,23 @@ class PyHocaCLI(x2go.X2GoClient):
             elif self.args.terminate:
                 self.terminate_session(self.x2go_session_hash)
 
+            elif self.args.new and self.args.try_resume and self.try_resume_session(self.x2go_session_hash):
+                # nothing to do here, if resumption attempt was successful
+                pass
+
             elif self.args.new:
                 self.new_session(self.x2go_session_hash)
 
             if not (self.args.new or self.args.resume or self.args.share_desktop or self.args.session_profile):
                 sys.exit(0)
-
-            if self.args.resume or self.args.new:
-                self._pyhoca_logger("give the X2Go session some time to come up...", loglevel=x2go.loglevel_NOTICE, )
-
             i=0
             while 0 < self.get_session(self.x2go_session_hash).get_progress_status() < 100:
-                time.sleep(1)
+                sleep(1)
                 i+=1
 
+            session = self._X2GoClient__get_session(self.x2go_session_hash)
+            session.set_master_session()
+
             if self._X2GoClient__session_ok(self.x2go_session_hash):
 
                 profile_name = self._X2GoClient__get_session_profile_name(self.x2go_session_hash)
@@ -567,11 +746,11 @@ class PyHocaCLI(x2go.X2GoClient):
                     session_duration = 0
 
                     while self._X2GoClient__session_ok(self.x2go_session_hash):
-                        time.sleep(2)
+                        sleep(2)
                         session_duration +=2
 
                     # wait a little while before telling the user what has happened
-                    time.sleep(2)
+                    sleep(2)
 
                     # refresh session status so we can be most accurate on what we report below
                     self._X2GoClient__list_sessions(self.x2go_session_hash)
@@ -588,21 +767,21 @@ class PyHocaCLI(x2go.X2GoClient):
                     if self.args.share_desktop:
                         self._pyhoca_logger("Terminating X2Go shared desktop session %s" % session_name, loglevel=x2go.loglevel_INFO, )
                         self._X2GoClient__terminate_session(self.x2go_session_hash)
-                        time.sleep(2)
+                        sleep(2)
                         self._pyhoca_logger("X2Go session %s has been terminated" % session_name, loglevel=x2go.loglevel_NOTICE, )
                     elif self.args.terminate_on_ctrl_c:
                         self._pyhoca_logger("Terminating X2Go session %s" % session_name, loglevel=x2go.loglevel_INFO, )
                         self._X2GoClient__terminate_session(self.x2go_session_hash)
                         # giving nxproxy's SSH tunnel some time to settle
-                        time.sleep(2)
+                        sleep(2)
                         self._pyhoca_logger("X2Go session %s has been terminated" % session_name, loglevel=x2go.loglevel_NOTICE, )
                     else:
                         self._pyhoca_logger("Suspending X2Go session %s" % session_name, loglevel=x2go.loglevel_INFO, )
                         self._X2GoClient__suspend_session(self.x2go_session_hash)
                         # giving nxproxy's SSH tunnel some time to settle
-                        time.sleep(2)
+                        sleep(2)
                         self._pyhoca_logger("X2Go session %s has been suspended" % session_name, loglevel=x2go.loglevel_NOTICE, )
 
-        except x2go.X2GoSessionException, e:
+        except x2go.X2GoSessionException as e:
             self._pyhoca_logger("X2GoSessionException occured:", loglevel=x2go.loglevel_ERROR)
             self._pyhoca_logger("-> %s" % str(e), loglevel=x2go.loglevel_ERROR)
